# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""
Wrapper around various loggers and progress bars (e.g., tqdm).
"""
from collections import OrderedDict
from contextlib import contextmanager
import json
import logging
import os
import sys
from numbers import Number
from typing import Dict, Optional

import torch

from .meters import AverageMeter, StopwatchMeter, TimeMeter


logger = logging.getLogger(__name__)


def progress_bar(
    iterator,
    log_format: Optional[str] = None,
    log_interval: int = 100,
    epoch: Optional[int] = None,
    prefix: Optional[str] = None,
    tensorboard_logdir: Optional[str] = None,
    default_log_format: str = 'tqdm',
    mlflow_server_url: Optional[str] = None,
    mlflow_experiment_name: str = 'test-experiment',
    mlflow_run_id: Optional[str] = None,
):
    if log_format is None:
        log_format = default_log_format
    if log_format == 'tqdm' and not sys.stderr.isatty():
        log_format = 'simple'

    bar: BaseProgressBar
    if log_format == 'json':
        bar = JsonProgressBar(iterator, epoch, prefix, log_interval)
    elif log_format == 'none':
        bar = NoopProgressBar(iterator, epoch, prefix)
    elif log_format == 'simple':
        bar = SimpleProgressBar(iterator, epoch, prefix, log_interval)
    elif log_format == 'tqdm':
        bar = TqdmProgressBar(iterator, epoch, prefix)
    else:
        raise ValueError('Unknown log format: {}'.format(log_format))

    if tensorboard_logdir:
        try:
            # [FB only] custom wrapper for TensorBoard
            import palaas  # noqa
            from .fb_tbmf_wrapper import FbTbmfWrapper
            bar = FbTbmfWrapper(bar, log_interval)
        except ImportError:
            bar = TensorboardProgressBarWrapper(bar, tensorboard_logdir)

    if mlflow_server_url:
        bar = MlflowLogWrapper(bar, mlflow_server_url, mlflow_experiment_name,
                               mlflow_run_id, log_interval, mlflow_value_length_limit=250)
        logger.info(f'MLFlow will be updated {len(iterator)/log_interval} '
                     f'times during this epoch (or validation session)')
    return bar


def build_progress_bar(
    args,
    iterator,
    epoch: Optional[int] = None,
    prefix: Optional[str] = None,
    default: str = 'tqdm',
    no_progress_bar: str = 'none',
):
    """Legacy wrapper that takes an configargparse.Namespace."""
    if getattr(args, 'no_progress_bar', False):
        default = no_progress_bar
    if getattr(args, 'distributed_rank', 0) == 0:
        tensorboard_logdir = getattr(args, 'tensorboard_logdir', None)
    else:
        tensorboard_logdir = None
    return progress_bar(
        iterator,
        log_format=args.log_format,
        log_interval=args.log_interval,
        epoch=epoch,
        prefix=prefix,
        tensorboard_logdir=tensorboard_logdir,
        default_log_format=default,
        mlflow_server_url=getattr(args, 'mlflow_server_url', None),
        mlflow_experiment_name=getattr(args, 'mlflow_experiment_name', None),
        mlflow_run_id=getattr(args, 'mlflow_run_id', None),
    )


def format_stat(stat):
    if isinstance(stat, Number):
        stat = '{:g}'.format(stat)
    elif isinstance(stat, AverageMeter):
        stat = '{:.3f}'.format(stat.avg)
    elif isinstance(stat, TimeMeter):
        stat = '{:g}'.format(round(stat.avg))
    elif isinstance(stat, StopwatchMeter):
        stat = '{:g}'.format(round(stat.sum))
    elif torch.is_tensor(stat):
        stat = stat.tolist()
    return stat


class BaseProgressBar(object):
    """Abstract class for progress bars."""
    def __init__(self, iterable, epoch=None, prefix=None):
        self.iterable = iterable
        self.offset = getattr(iterable, 'offset', 0)
        self.epoch = epoch
        self.prefix = ''
        if epoch is not None:
            self.prefix += 'epoch {:03d}'.format(epoch)
        if prefix is not None:
            self.prefix += ' | {}'.format(prefix)

    def __len__(self):
        return len(self.iterable)

    def __enter__(self):
        return self

    def __exit__(self, *exc):
        return False

    def __iter__(self):
        raise NotImplementedError

    def log(self, stats, tag=None, step=None):
        """Log intermediate stats according to log_interval."""
        raise NotImplementedError

    def print(self, stats, tag=None, step=None):
        """Print end-of-epoch stats."""
        raise NotImplementedError

    def _str_commas(self, stats):
        return ', '.join(key + '=' + stats[key].strip()
                         for key in stats.keys())

    def _str_pipes(self, stats):
        return ' | '.join(key + ' ' + stats[key].strip()
                          for key in stats.keys())

    def _format_stats(self, stats):
        postfix = OrderedDict(stats)
        # Preprocess stats according to datatype
        for key in postfix.keys():
            postfix[key] = str(format_stat(postfix[key]))
        return postfix


@contextmanager
def rename_logger(logger, new_name):
    old_name = logger.name
    if new_name is not None:
        logger.name = new_name
    yield logger
    logger.name = old_name


class JsonProgressBar(BaseProgressBar):
    """Log output in JSON format."""

    def __init__(self, iterable, epoch=None, prefix=None, log_interval=1000):
        super().__init__(iterable, epoch, prefix)
        self.log_interval = log_interval
        self.stats = None
        self.tag = None

    def __iter__(self):
        size = float(len(self.iterable))
        for i, obj in enumerate(self.iterable, start=self.offset):
            yield obj
            if (
                self.stats is not None
                and i > 0
                and self.log_interval is not None
                and (i + 1) % self.log_interval == 0
            ):
                update = (
                    self.epoch - 1 + float(i / size)
                    if self.epoch is not None
                    else None
                )
                stats = self._format_stats(self.stats, epoch=self.epoch, update=update)
                with rename_logger(logger, self.tag):
                    logger.info(json.dumps(stats))

    def log(self, stats, tag=None, step=None):
        """Log intermediate stats according to log_interval."""
        self.stats = stats
        self.tag = tag

    def print(self, stats, tag=None, step=None):
        """Print end-of-epoch stats."""
        self.stats = stats
        if tag is not None:
            self.stats = OrderedDict([(tag + '_' + k, v) for k, v in self.stats.items()])
        stats = self._format_stats(self.stats, epoch=self.epoch)
        with rename_logger(logger, tag):
            logger.info(json.dumps(stats))

    def _format_stats(self, stats, epoch=None, update=None):
        postfix = OrderedDict()
        if epoch is not None:
            postfix['epoch'] = epoch
        if update is not None:
            postfix['update'] = round(update, 3)
        # Preprocess stats according to datatype
        for key in stats.keys():
            postfix[key] = format_stat(stats[key])
        return postfix


class NoopProgressBar(BaseProgressBar):
    """No logging."""

    def __init__(self, iterable, epoch=None, prefix=None):
        super().__init__(iterable, epoch, prefix)

    def __iter__(self):
        for obj in self.iterable:
            yield obj

    def log(self, stats, tag=None, step=None):
        """Log intermediate stats according to log_interval."""
        pass

    def print(self, stats, tag=None, step=None):
        """Print end-of-epoch stats."""
        pass


class SimpleProgressBar(BaseProgressBar):
    """A minimal logger for non-TTY environments."""

    def __init__(self, iterable, epoch=None, prefix=None, log_interval=1000):
        super().__init__(iterable, epoch, prefix)
        self.log_interval = log_interval
        self.stats = None
        self.tag = None

    def __iter__(self):
        for i, obj in enumerate(self.iterable, start=self.offset):
            yield obj
            if (
                self.stats is not None
                and i > 0
                and self.log_interval is not None
                and (i + 1) % self.log_interval == 0
            ):
                postfix = self._str_commas(self.stats)
                with rename_logger(logger, self.tag):
                    logger.info('{}:  {:5d} {}'.format(self.prefix, i, postfix))

    def log(self, stats, tag=None, step=None):
        """Log intermediate stats according to log_interval."""
        self.stats = self._format_stats(stats)
        self.tag = tag

    def print(self, stats, tag=None, step=None):
        """Print end-of-epoch stats."""
        postfix = self._str_pipes(self._format_stats(stats))
        with rename_logger(logger, tag):
            logger.info('{} | {}'.format(self.prefix, postfix))


class TqdmProgressBar(BaseProgressBar):
    """Log to tqdm."""

    def __init__(self, iterable, epoch=None, prefix=None):
        super().__init__(iterable, epoch, prefix)
        from tqdm import tqdm
        self.tqdm = tqdm(iterable, self.prefix, leave=False)

    def __iter__(self):
        return iter(self.tqdm)

    def log(self, stats, tag=None, step=None):
        """Log intermediate stats according to log_interval."""
        self.tqdm.set_postfix(self._format_stats(stats), refresh=False)

    def print(self, stats, tag=None, step=None):
        """Print end-of-epoch stats."""
        postfix = self._str_pipes(self._format_stats(stats))
        self.tqdm.write('{} | {}'.format(self.tqdm.desc, postfix))


try:
    from tensorboardX import SummaryWriter
    _tensorboard_writers = {}
except ImportError:
    SummaryWriter = None


class TensorboardProgressBarWrapper(BaseProgressBar):
    """Log to tensorboard."""

    def __init__(self, wrapped_bar, tensorboard_logdir):
        self.wrapped_bar = wrapped_bar
        self.tensorboard_logdir = tensorboard_logdir

        if SummaryWriter is None:
            logger.warning(
                "tensorboard or required dependencies not found, please see README "
                "for using tensorboard. (e.g. pip install tensorboardX)"
            )

    def _writer(self, key):
        if SummaryWriter is None:
            return None
        _writers = _tensorboard_writers
        if key not in _writers:
            _writers[key] = SummaryWriter(os.path.join(self.tensorboard_logdir, key))
            _writers[key].add_text('sys.argv', " ".join(sys.argv))
        return _writers[key]

    def __iter__(self):
        return iter(self.wrapped_bar)

    def log(self, stats, tag=None, step=None):
        """Log intermediate stats to tensorboard."""
        self._log_to_tensorboard(stats, tag, step)
        self.wrapped_bar.log(stats, tag=tag, step=step)

    def print(self, stats, tag=None, step=None):
        """Print end-of-epoch stats."""
        self._log_to_tensorboard(stats, tag, step)
        self.wrapped_bar.print(stats, tag=tag, step=step)

    def _log_to_tensorboard(self, stats, tag=None, step=None):
        writer = self._writer(tag or '')
        if writer is None:
            return
        if step is None:
            step = stats['num_updates']
        for key in stats.keys() - {'num_updates'}:
            if isinstance(stats[key], AverageMeter):
                writer.add_scalar(key, stats[key].val, step)
            elif isinstance(stats[key], Number):
                writer.add_scalar(key, stats[key], step)


class MlflowLogWrapper(BaseProgressBar):
    """Log to ML Flow."""

    def __init__(
        self,
        wrapped_bar: BaseProgressBar,
        mlflow_server_url: str,
        mlflow_experiment_name: str = 'test-experiment',
        mlflow_run_id: Optional[str] = None,
        log_interval:int = 100,
        mlflow_value_length_limit: str = 250,
    ):
        self.wrapped_bar = wrapped_bar
        self.server_url = mlflow_server_url
        self.mlflow_value_length_limit = mlflow_value_length_limit
        self.log_interval = log_interval
        try:
            from mlflow.tracking import MlflowClient
            self.client: MlflowClient = MlflowClient(tracking_uri=self.server_url)
        except ImportError:
            logger.warning("mlflow or required dependencies not found, "
                           "please see README for using mlflow. (e.g. pip install mlflow)")
            self.client = None

        if mlflow_run_id is not None:
            self._run_id = mlflow_run_id
        else:
            experiment = self.client.get_experiment_by_name(mlflow_experiment_name)
            if experiment is None:
                exp_id = self.client.create_experiment(mlflow_experiment_name)
            else:
                exp_id = experiment.experiment_id

            logger.info('Creating a new run in %s experiment on ML-Flow.', mlflow_experiment_name)
            self._run_id = self.client.create_run(exp_id).info.run_uuid
            logger.info('Created run id: %s', mlflow_experiment_name)

    def __iter__(self):
        return iter(self.wrapped_bar)

    @property
    def run_id(self):
        return self._run_id

    def log(self, stats, tag='', step: int = None):
        """Log intermediate stats to mlflow."""
        if step is None and 'num_updates' in stats:
            step = stats['num_updates']
        if self.log_interval is not None and (step + 1) % self.log_interval == 0:
            self._log_to_mlflow(stats, tag, step)
        self.wrapped_bar.log(stats, tag=tag, step=step)

    def print(self, stats, tag: str = '', step: int = None):
        """Print end-of-epoch stats."""
        self._log_to_mlflow(stats, tag, step)
        self.wrapped_bar.print(stats, tag=tag, step=step)

    def __exit__(self, *exc):
        self.client.set_terminated(self.run_id)
        return False

    def _log_to_mlflow(self, stats, tag='', step=None):
        for key in stats.keys() - {'num_updates'}:
            value = None
            if isinstance(stats[key], AverageMeter):
                value = stats[key].val
            elif isinstance(stats[key], Number):
                value = stats[key]
            if value is not None:
                if tag != '':
                    key = f'{tag}_{key}'
                self.client.log_metric(run_id=self.run_id, key=key,
                                       value=self._trim_value_to_limit(value), step=step)

    def log_training_params(self, params: Dict[str, str]):
        for key, value in params.items():
            if not key.startswith('mlflow'):
                self.client.log_param(run_id=self.run_id, key=key,
                                      value=self._trim_value_to_limit(value))

    def _trim_value_to_limit(self, value):
        if isinstance(value, str):
            return value[:self.mlflow_value_length_limit]
        return value